/**
 # Copyright (c) 2015 Riverbed Technology, Inc.
 #
 # This software is licensed under the terms and conditions of the MIT License
 # accompanying the software ("License").  This software is distributed "AS IS"
 # as set forth in the License.
 */

body
{
   height:100%;
   max-height:100%;
   overflow:auto;
   position: absolute;
   top: 0;
   left: 0;
   padding:0;
   margin:0;
   border:0;

   font-family: Helvetica, clean, sans-serif;
   font-size: 11pt;
   background-color: #fff;
}

.header
{
  float: top;
  margin: 10px;
  padding-left: 10px;
  background-color: #ccc;
  border: 1px solid #999;
  height: 65px;
  width: 1160px;
  }

.headerleft
{
  margin-top: 13px;
  padding-left: 25px;
  float: left;
  }

.headertitle
{
  font-size: 22pt;
  }

.breadcrumbs
{
  font-size: 10pt;
  font-style: italic;
  padding-bottom: 6px;
  }

.breadcrumbs a
{
  color: #000;
  }

.headerright
{
  vertical-align: bottom;
  font-size: 10pt;
  font-style: italic;
  float: right;
  padding-right: 8px;
  padding-top: 45px;
  }

.headerimg
{
  float: left;
  overflow: hidden;
  vertical-align: middle;
  margin-top: 10px;
  width:48px;
  height:48px;
  background-image:url(data:image/png;base64,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)  }

.headerimg {
}

.main
{
   position: fixed;
   top: 90px;
   bottom: 30px;
}

.main-printable
{
  height: 100%;
}

.footer
{
  position: absolute;
  bottom: 0;
  clear:both;
  font-size: 80%;
  font-style: italic;
  color: #666;
  border-top: 1px solid #bbb;
  text-align:center;
  padding-top: 0.2em;
  padding-bottom: 0.2em;
  margin-left: 10px;
  margin-right: 10px;
  margin-top: 10px;
  width: 1172px;
  }

.navbar
{
  clear: left;
  float: left;
  display: block;
  width: 250px;
  margin-left: 10px;
  margin-right: 10px;
  margin-top: 0px;
  overflow: auto;
  height: 100%;
  font-size: 95%;
  border: 1px solid #999;
  }

.toc, .links
{
  background-color: #ccc;
  color: #000;
  margin-left: 0em;
  margin-right: 0px;
  margin-bottom: 1px;
  padding-top: 1px;
  padding-bottom: 1px;
  }

.toc a, .links a
{
  color: #000;
  }

.toctitle
{
  font-family: Helvetica, clean, sans-serif;
  font-weight: bold;
  padding-left:0.5em;
  color: #000;
  }

.toc > p, .links > p
{
  padding-left:0;
  }

.toc > ul,
.toc > ul > li > ul,
.toc > ul > li > ul > li > ul,
.toc > ul > li > ul > li > ul > li > ul,
.links > ul,
.links > ul > li> ul,
.links > ul > li > ul > li > ul
{
  font-size: 95%;
  list-style: disc;
  padding-left: 1.5em;
  }


.content
{
  border: 1px solid #bbb;
  position: absolute;
  left: 270px;
  margin-right:5px;
  padding-top:0px;
  padding-right:5px;
  width: 905px;
  height: 100%;

  background-color: #ffffff;
  overflow: auto;
  }

.content-printable
{
  margin-top: 0px;
  margin-left: 0px;
  margin-right:0px;
  padding-top:0;
  max-width: 950px;
  background-color: #ffffff;
}

h1, h2, h3, h4, span.h4, h5, span.h5
{
  color: #111111;
  font-weight: 400;
  padding-left: 0.5em;
  padding-top: 0.4em;
  margin-bottom: 0.3em;
  }

h1
{
  font-size: 170%;
  background-color: #ddd;
  padding-top: 0.4em;
  padding-bottom: 0.2em;
  border-top: 1px solid #CCC;
  border-bottom: 1px solid #CCC;
  margin-top: 0.5em;
  margin-bottom: 0.5em;
  }

.break
{
  page-break-before: always;
}

h2
{
  font-size: 150%;
  font-weight: bold;

  margin: 24px 0 6px;

  border-top: 1px solid #aaa;
  padding-top: 0.4em;
  padding-bottom: 0.2em;
  margin-bottom: 0.3em;
  }

h3
{
  font-size: 130%;
  font-weight: bold;
  border-top: 1px solid #ccc;
  padding-bottom: 0.2em;
  }

h4, span.h4
{
  font-size: 120%;
  font-weight: bold;
  }

span.h4, span.h5
{
  display: block;
  clear: both;
}

h5, span.h5
{
  font-size: 110%;
  }

p
{
  margin-top: 0px;
  margin-left: 20px;
  padding-right: 2em;
  }
a
{
  /*color: #0099ff;*/
  color: #FF8040;

  margin: 0;
  padding: 0;
  vertical-align: baseline;
  }
a:hover
{
  text-decoration: none;
  color: #ff6600;
  }

dt
{
  font-size: 105%;
  font-weight: bold;
  padding-left: 0.5em;
  padding-top: 0.6em;
  padding-bottom: 0.2em;
  }

li
{
  line-height: 140%;
  }


pre
{
  margin-left: 20px;
  padding: 0px 24px;
  white-space: pre-wrap;
  font-size: 85%;

  background-color: #ebebff;
  border: 1px dashed #a95;
  overflow: auto;
  padding: 6px 10px;
  }

code
{
  font-family: Consolas, Monaco, Andale Mono, monospace;
  font-size: 90%;
  line-height: 1.5;
  }

table.paramtable
{
  font-family: verdana,arial,sans-serif;
  font-size:11px;
  color:#333333;
  margin-left: 20px;
  border-width: 1px;
  border-color: #666666;
  border-collapse: collapse;
  }

table.paramtable col.paramtable-propname
{
    width: 20%;
}

table.paramtable col.paramtable-proptype
{
    width: 12%;
}

table.paramtable col.paramtable-description
{
    width: 38%;
}

table.paramtable col.paramtable-notes
{
    width: 30%;
}

table.paramtable th
{
  border-width: 1px;
  padding: 4px;
  border-style: solid;
  border-color: #666666;
  background-color: #dedede;
  }

table.paramtable td
{
  border-width: 1px;
  padding: 4px;
  border-style: solid;
  border-color: #666666;
  background-color: #ffffff;
  }

pre.servicedef
{
  margin-left: 10px;
  background-color: #ffd;
 }

span.servicedef-type, a.servicedef-type

{
  color: #070;
  font-style: italic;
 }

span.servicedef-property
{
  color: #009;
  }

span.servicedef-indent
{
    padding-left: 1em;
}

span.servicedef-basename
{
  color: #666;
  font-style: italic;
  }

span.xmlschema-type, a.xmlschema-type
{
  color: #070;
  font-style: italic;
 }

span.xmlschema-attribute
{
  color: #099;
  }

span.xmlschema-element
{
  color: #009;
  }

span.content-attribute
{
  font-family: Consolas, Monaco, Andale Mono, monospace;
}

.tabContainer
{
  margin-left: 20px;
  }

.tabContainer .digiTabs
{
  list-style:none;
  display:block;
  overflow:hidden;
  margin:0;
  padding:0px;
  position:relative;
  top:1px;
  }
.tabContainer .digiTabs li
{
  float:left;
  background-color:#fff;
  border:1px solid #e1e1e1;
  padding:5px!important;
  cursor:pointer;
  border-bottom:1px solid #e1e1e1;
  margin-right:10px;
  font-family:verdana;
  font-size:.8em;
  font-weight:bold;
  color:#bbb;
  }
.tabContainer .digiTabs .selected
{
  background-color:#fff;
  color:#393939;
  border-bottom:1px solid #fff;
  }

.tabContent
{
  padding-right:10px;
  background-color:#fff;
  overflow:hidden;
  float:left;
  margin-bottom:10px;
  border:1px solid #e1e1e1;
  width:850px;
  }
